@@ -1,13 +1,16 @@
-# gSOBI algorithm (combination of SOBI and vSOBI)
-gSOBI <- function (X, k1 = 1:12, k2 = 1:3, b = 0.9, eps = 1e-06, maxiter = 1000, 
-                   ordered = FALSE, acfk = NULL, original = TRUE, alpha = 0.05, ...){
-  MEAN <- colMeans(X)
-  COV <- cov(X)
-  EVD <- eigen(COV, symmetric = TRUE)
-  COV.sqrt.i <- EVD$vectors %*% tcrossprod(diag(EVD$values^(-0.5)), EVD$vectors)
-  X.C <- sweep(X, 2, MEAN, "-")
-  Y <- tcrossprod(X.C, COV.sqrt.i)
+# Method gSOBI
+gSOBI <- function(X, ...) UseMethod("gSOBI")
+
+# gSOBI function for PVC (combination of SOBI and vSOBI)
+gSOBI.default <- function(X, k1 = 1:12, k2 = 1:3, b = 0.9, eps = 1e-06, maxiter = 1000, 
+                   ordered = FALSE, acfk = NULL, original = TRUE, alpha = 0.05, ...) {
+  if (!is.numeric(X)) stop("non-numeric data")
+  if (any(is.na(X) | is.infinite(X))) stop("missing/infinite values are not allowed")
+  n <- nrow(X)
   p <- ncol(X)
+  prep <- .Call("PREPBSS", X, n, PACKAGE = "tsBSS") #calling the function PREPBSS
+  Y <- prep$Y 
+  
   U <- diag(p) #Initial value for the orthogonal matrix U
   crit <- Inf
   iter <- 0
@@ -15,24 +18,24 @@ gSOBI <- function (X, k1 = 1:12, k2 = 1:3, b = 0.9, eps = 1e-06, maxiter = 1000,
   K2 <- length(k2)
   Tk1 <- array(NA, dim = c(p, p, K1))
   Tk2 <- array(NA, dim = c(p, p, K2))
-  while(crit > eps) {
-      for (i in 1:K1) {
-        Tk1[ , , i] <- TIK1c(Y, U, k = k1[i])
-      }
-      for (i in 1:K2) {
-        Tk2[ , , i] <- TIKc(Y, U, k = k2[i], method = 3)
-      }
+  while (crit > eps) {
+    for (i in 1:K1) {
+      Tk1[ , , i] <- .Call("TIK1", Y, U, k1[i], PACKAGE = "tsBSS")$Tik
+    }
+    for (i in 1:K2) {
+      Tk2[ , , i] <- .Call( "TIK", Y, U, k2[i], method = 3, PACKAGE = "tsBSS")$Tik
+    }
     TU <- b*apply(Tk1, c(1, 2), sum) + (1 - b)*apply(Tk2, c(1, 2), sum)
-    EVDt <- eigen(tcrossprod(TU), symmetric = TRUE)
-    COVt.sqrt.i <- EVDt$vectors %*% tcrossprod(diag(EVDt$values^(-0.5)), EVDt$vectors)
+    EVDt <- .Call("EIGEN", tcrossprod(TU), PACKAGE = "tsBSS")
+    COVt.sqrt.i <- EVDt$vectors %*% tcrossprod(diag(as.vector(EVDt$values)^(-0.5)), EVDt$vectors)
     U.new <- COVt.sqrt.i %*% TU #Updated U
     crit <- sqrt(sum((abs(U.new) - abs(U))^2)) #Comparing the current and the new matrix U
     iter <- iter + 1
-    if(iter > maxiter) stop("maxiter reached without convergence")
+    if (iter > maxiter) stop("maxiter reached without convergence")
     U <- U.new
-  } #While the criterion value is below tolerance value.
-  W <- crossprod(U, COV.sqrt.i) #Unmixing matrix
-  S <- tcrossprod(X.C, W)
+  } #While the criterion value is below tolerance value
+  W <- crossprod(U, prep$COV.sqrt.i) #Unmixing matrix
+  S <- tcrossprod(prep$X.C, W)
   if (ordered == TRUE) { #Ordering by volatility
     if (is.null(acfk) == TRUE) { acfk <- 1:max(k1, k2) }
     ord <- ordf(S, acfk, p, W, alpha, ...)
@@ -43,12 +46,10 @@ gSOBI <- function (X, k1 = 1:12, k2 = 1:3, b = 0.9, eps = 1e-06, maxiter = 1000,
       S <- ord$RS # Residuals based on ARMA fit, if applicable; otherwise original IC's
       Sraw <- ord$S
       Sraw <- ts(Sraw, names = paste("Series", 1:p))
-      if (is.ts(X)) attr(Sraw, "tsp") <- attr(X, "tsp")
     }
   }
   S <- ts(S, names = paste("Series", 1:p))
-  if (is.ts(X)) attr(S, "tsp") <- attr(X, "tsp")
-  RES <- list(W = W, k1 = k1, k2 = k2, S = S, MU = MEAN)
+  RES <- list(W = W, k1 = k1, k2 = k2, S = S, MU = prep$MEAN)
   if (ordered == TRUE) {
     if (original == FALSE) {
       RES$Sraw <- Sraw
@@ -64,6 +65,50 @@ gSOBI <- function (X, k1 = 1:12, k2 = 1:3, b = 0.9, eps = 1e-06, maxiter = 1000,
   RES
 }
 
+gSOBI.ts <- function(X, ...) {
+  x <- as.matrix(X)
+  RES <- gSOBI.default(x, ...)
+  S <- RES$S
+  attr(S, "tsp") <- attr(X, "tsp")
+  RES$S <- S
+  if (!is.null(RES$Sraw)) {
+    Sraw <- RES$Sraw
+    attr(Sraw, "tsp") <- attr(X, "tsp")
+    RES$Sraw <- Sraw
+  }
+  RES
+}
+
+gSOBI.xts <- function(X, ...) {
+  x <- as.matrix(X)
+  RES <- gSOBI.default(x, ...)
+  S <- xts::as.xts(RES$S)
+  attr(S, "index") <- attr(X, "index")
+  xts::xtsAttributes(S) <- xts::xtsAttributes(X) #attributes additional to zoo
+  RES$S <- S
+  if (!is.null(RES$Sraw)) {
+    Sraw <- xts::as.xts(RES$Sraw)
+    attr(Sraw, "index") <- attr(X, "index")
+    xts::xtsAttributes(Sraw) <- xts::xtsAttributes(X)
+    RES$Sraw <- Sraw
+  }
+  RES
+}
+
+gSOBI.zoo <- function(X, ...) {
+  x <- as.matrix(X)
+  RES <- gSOBI.default(x, ...)
+  S <- zoo::as.zoo(RES$S)
+  attr(S, "index") <- attr(X, "index")
+  RES$S <- S
+  if (!is.null(RES$Sraw)) {
+    Sraw <- zoo::as.zoo(RES$Sraw)
+    attr(Sraw, "index") <- attr(X, "index")
+    RES$Sraw <- Sraw
+  }
+  RES
+}
+
 #Function to order the components (calculates the volatilities of the components or their residuals)
 # Used by the functions gFOBI, gJADE, gSOBI, vSOBI, PVC and FixNA
 ordf <- function(S, acfk, p, W, alpha, ...) {
@@ -74,7 +119,7 @@ ordf <- function(S, acfk, p, W, alpha, ...) {
   S2 <- S
   coef <- numeric(p)
   for (i in 1:p) {
-    if(lineff[i] < alpha) {
+    if (lineff[i] < alpha) {
       fits[[i]] <- forecast::auto.arima(S[, i], stationary = T, seasonal = F, ...)
       S2[, i] <- fits[[i]]$residuals #Replaces series with residuals (if autocorrelation exists)
       armaeff[i] <- 1 #Logical vector: 1 if series is replaced with residuals; 0 if not
@@ -85,16 +130,16 @@ ordf <- function(S, acfk, p, W, alpha, ...) {
   fits2 <- vector("list", p)
   j <- 1
   ordered <- order(ord, decreasing = T)
-  for(i in ordered) {
+  for (i in ordered) {
     fits2[[j]] <- fits[[i]]
-    j = j + 1
+    j <- j + 1
   }
   list(S = S[, ordered],
        RS = S2[, ordered], #Residuals if ARMA effects; otherwise original independent component
        W = W[ordered, ],
        fits = fits2,
        volTS = vol$TS[ordered], volP = vol$p_val[ordered],
-       armaeff = armaeff[ordered], linTS = lblin1$TS[ordered], linP = lineff[ordered])
+       armaeff = armaeff[ordered] == 1, linTS = lblin1$TS[ordered], linP = lineff[ordered])
 }
 
 
@@ -102,3 +147,16 @@ ordf <- function(S, acfk, p, W, alpha, ...) {
   print.listof(x[(names(x) != "S") & (names(x) != "Sraw") & (names(x) != "MU") & (names(x) != "fits") & (names(x) != "armaeff") & (names(x) != "linTS")
                  & (names(x) != "linP") & (names(x) != "volTS") & (names(x) != "volP")], ...)
 }
+
+`plot.bssvol` <- function(x, ...) {
+  S <- x$S
+  if(ncol(S) <= 2) {
+    plot(S, ...)
+    } else {
+      if (any(class(S) %in% c("mts", "xts", "zoo"))) {
+        plot(S, ...)
+        } else {
+          pairs(S, ...)
+        }
+    }
+}
